#!/usr/bin/python3.5
#
# This is the Python module implementing access to the Hitachi hd44780
# 2-line LCD display
# U. Raich UCC, Nov.2017
import sys,time
import pigpio

class LcdDisplayClass:
    I2C_BUS                =    1
    I2C_FLAGS              =    0
    HD44780_ADDRESS        = 0x3f
    
    HD44780_CLEAR          = 0x01
    HD44780_HOME           = 0x02
    HD44780_ENTRY_MODE     = 0x04
    HD44780_CURSOR_DIR     = 0x02
    HD44780_CURSOR_SHIFT   = 0x10
    HD44780_DISPLAY_CURSOR = 0x08
    HD44780_DISPLAY_ON     = 0x04
    HD44780_DISPLAY_OFF    = 0x00
    HD44780_CURSOR_ON      = 0x02
    HD44780_CURSOR_OFF     = 0x00
    HD44780_CURSOR_BLINK   = 0x01
    HD44780_CURSOR_NOBLINK = 0x00

    HD44780_ENTIRE_DISPLAY = 0x04
    HD44780_CURSOR_ONOFF   = 0x02
    HD44780_BLINK          = 0x01
    HD44780_CURSOR         = 0x10
    HD44780_1_LINE         = 0x00
    DH44780_CURSOR_MOVE    = 0x00
    DH44780_DISPLAY_SHIFT  = 0x08
    DH44780_SHIFT_LEFT     = 0x00
    DH44780_SHIFT_RIGHT    = 0x04

    HD44780_FUNCTION       = 0x20
    
    HD44780_8_BIT_MODE     = 0x10
    HD44780_4_BIT_MODE     = 0x00
    HD44780_TWO_LINES      = 0x08
    HD44780_5x10_FONT      = 0x04
    HD44780_5x8_FONT       = 0x00
    HD44780_1_LINE         = 0x00
    HD44780_2_LINE         = 0x08
    HD44780_CURSOR_DEC     = 0x00
    HD44780_CURSOR_INC     = 0x02
    HD44780_DISPLAY_SHIFT  = 0x01
    
    HD44780_CGRAM_AHDR     = 0x10
    HD44780_HDRAM_AHDR     = 0x20
    HD44780_BUSY_FLAG      = 0x40

    HD44780_BL             = 0x08
    HD44780_BL_SHIFT       = 0x03

    HD44780_EN             = 0x04
    HD44780_READ           = 0x02
    HD44780_WRITE          = 0x00
    HD44780_CMD_REG        = 0x00
    HD44780_DATA_REG       = 0x01

    HD44780_SET_DDRAM      = 0x80
    HD44780_SET_CGRAM      = 0x40
    HD44780_2ND_LINE       = 0x40

    HD44780_INCREMENT      = True
    HD44780_DECREMENT      = False
    HD44780_ON             = True
    HD44780_OFF            = False

    HD44780_SUCCESS        = 0
    HD44780_MAX_STRLEN     = 16


# pinout for 20x2 LCD via PCF8574AN:
# ----------
# 0x80    P7 -  - D7
# 0x40    P6 -  - D6
# 0x20    P5 -  - D5
# 0x10    P4 -  - D4
# -----------
# 0x08    P3 -  - BL   Backlight (LCD has 3 RGB LEDs - common anode)
# 0x04    P2 -  - EN   Starts Data read/write
# 0x02    P1 -  - RW   low: write, high: read
# 0x01    P0 -  - RS   Register Select: 0: Instruction Register (IR) (AC when read), 
#                                       1: data register (DR)

    pi        = -1
    backlight =  1                      # backlight is on by default
    handle    = -1

    def __init__(self):
        global pi
#
# open the connection to the pigpiod daemon
#
        pi = pigpio.pi()
        if not pi.connected:
            print("Could not connect to the pigpiod daemon, giving up...")
            exit()
#
# open the I2C bus driver
#
        handle = pi.i2c_open(self.I2C_BUS,
                             self.HD44780_ADDRESS,self.I2C_FLAGS)
        if handle < 0:
            print("Could not open the I2C bus for address ",
                  self.HD44780_ADDRESS," giving up...")
            exit()
        else:
            self.handle = handle
#
# initialize the controller following the init procdure described in the data sheet
#
        cmd = self.HD44780_FUNCTION | self.HD44780_8_BIT_MODE | self.HD44780_BL
        for index in range(0,5):
#            print("in init: sending: 0x" + format(cmd,'02x')) 
            if pi.i2c_write_byte(handle,cmd) < 0:
                print("in init: Cannot write the I2C bus, ...giving up")
                exit
            self.__strobe(self.HD44780_CMD_REG)
            time.sleep(0.0042)      # sleep for 4.2 ms
#
# now initialize the hd44780 and set it to 4 bit mode
# this is still done in 8 bit mode and the command must be repeated
# in order to set 5x8 pixel characters in 4 bit mode
#
        cmd = self.HD44780_FUNCTION | self.HD44780_4_BIT_MODE | self.HD44780_BL;
#        print("in init: sending: 0x" + format(cmd,'02x')) 
        pi.i2c_write_byte(handle,cmd)
        self.__strobe(self.HD44780_CMD_REG)
#
# repeat the command, now in 4 bit mode
#
        cmd = self.HD44780_FUNCTION | self.HD44780_4_BIT_MODE | self.HD44780_5x8_FONT | self.HD44780_2_LINE;
#
# from now on use the class methods to write the command register
#
        retCode = self.__writeCmd(cmd)
        
        retCode = self.display(False)

        retCode = self.clear()

        retCode = self.home()

#        print("================== init: Initilization completed ==============================")

        retCode = self.readBusy()
#        print("in init: reading busy and address: ",retCode) 
        if (retCode & 0x80):
            print("in init: still busy")

        retCode = self.display(True)
        
    def close(self):
        global pi
        if self.handle > 0:
            pi.i2c_close(self.handle)
            self.handle = -1
        
#
# control and save the state of the backlight
#
    def setBacklight(self,onOff):
        global pi
        readBack = int(pi.i2c_read_byte(self.handle))
        if readBack < 0:
            print("Could not read I2C bus, giving up")
            return
        if (onOff):
            readBack |= self.HD44780_BL
            self.backlight = 1
        else:
            readBack &= ~self.HD44780_BL;
            self.backlight = 0;
        retCode = pi.i2c_write_byte(self.handle,readBack)
        if retCode < 0:
            printf("Could not write I2C bus")
            return
        
#
# local function to strobe data into the hd44780 controller
#
    def __strobe(self,register):
        global pi
        readBack = pi.i2c_read_byte(self.handle)
#        print("in strobe: readback: ", format(readBack,'04x'))
#        print("register: ", format(register,'04x'))
        if readBack < 0:
            print("in __strobe: Could not read I2C bus")
            return
 
        tmp = self.backlight << self.HD44780_BL_SHIFT
#        print("backlight: ",format(self.backlight,'02x'))
#        print("after shift: ",format(tmp,'02x'))
#        print("readback + backlight: ",format(readBack | self.backlight << self.HD44780_BL_SHIFT,'02x'))
        readBack = readBack | self.HD44780_EN | self.backlight << self.HD44780_BL_SHIFT | register;
#
# send the data with strobe set high
#
#        print("in strobe: sending: 0x" + format(readBack,'02x'))
        retCode = pi.i2c_write_byte(self.handle,readBack)
        if retCode < 0:
            print("in __strobe: Could not write I2C bus")
            return
        time.sleep(0.000037)        
#
# send the data with strobe set low
#
        readBack = (readBack | self.backlight << self.HD44780_BL_SHIFT | register) & ~self.HD44780_EN
#        print("in strobe: sending: 0x" + format(readBack,'02x'))        
        retCode = pi.i2c_write_byte(self.handle,readBack)
        time.sleep(0.000037)               
        if retCode < 0:
            print("in __strobe: Could not write I2C bus")
            return
       
#
# write a character to the lcd display
#
    def putC(self,character):
        global pi
#        print("in putC: writing character:",character)
        if self.handle < 0:
            printf("I2C bus is not initialized yet")
            return
#
# first treat the higher character nibble
#
#        print("H: ",format(ord(character),'04x'))
#        print("in putC: backlight: ",self.backlight)
        pcfc = (ord(character) &0xf0) | self.backlight << self.HD44780_BL_SHIFT | self.HD44780_WRITE |  self.HD44780_DATA_REG;
#        print("high nibble: ",format(pcfc,'02x'))
        retCode = pi.i2c_write_byte(self.handle,pcfc)
        self.__strobe(self.HD44780_DATA_REG)
        time.sleep(0.000037)         # wait for 37 us

#
# then the lower nibble
#
#        print("shifted by 4: ",format(pcfc,'04x'))
        pcfc = (ord(character) <<4) &0xf0 | self.backlight << self.HD44780_BL_SHIFT | self.HD44780_WRITE |  self.HD44780_DATA_REG;
#        print("low nibble: ",format(pcfc,'02x'))
        retCode = pi.i2c_write_byte(self.handle,pcfc)
        self.__strobe(self.HD44780_DATA_REG)
        time.sleep(0.000037)          # wait for 37 us

#
# write a string of characters to the display
#
    def putS(self,textString): 
        if len(textString) > self.HD44780_MAX_STRLEN:
            textString = textString[0:self.HD44780_MAX_STRLEN]
#        print ("in putS: the string to be put to the display: ",textString);
        for index in range(0,len(textString)):
            self.putC(textString[index])                         
#
# write a command to the command register
#
    def __writeCmd(self,cmd):
        global pi
        if self.handle < 0:
            print("I2C bus not initialized yet")
            return
#
# this is for the high nibble
#
        pcfCmd = (cmd &0xf0) | self.backlight << self.HD44780_BL_SHIFT | self.HD44780_WRITE | self.HD44780_CMD_REG;
#        print("writeCmd: sending high nibble: 0x",format(pcfCmd,'02x'))
        retCode = pi.i2c_write_byte(self.handle,pcfCmd)
        self.__strobe(self.HD44780_CMD_REG)
#
# and here for the low nibble
#
        pcfCmd = (cmd << 4) & 0xf0 | self.backlight << self.HD44780_BL_SHIFT | self.HD44780_WRITE | self.HD44780_CMD_REG;
#        print("writeCmd: sending low nibble: 0x",format(pcfCmd,'02x'))         
        retCode = pi.i2c_write_byte(self.handle,pcfCmd)
        self.__strobe(self.HD44780_CMD_REG)
        time.sleep(0.000037)          # wait for 37 us
        return retCode

#
# read a byte from the hd44780 command register in 4 bit mode
#  The high nibble is read first 
#
    def read(self,register):
        global pi
        if self.handle < 0:
            print("in read: the I2C bus is not initilized yet")
            return
        cmd =  0xf0 |self.backlight << self.HD44780_BL_SHIFT| self.HD44780_READ | register;
        retCode = pi.i2c_write_byte(self.handle,cmd)
        cmd |= self.HD44780_EN
#
# set enable high
#
        retCode = pi.i2c_write_byte(self.handle,cmd)
#
# read high nibble
#
        retCode = pi.i2c_read_byte(self.handle)
        if retCode < 0:
            print("in read: error when reading I2C bus")
            return
        else:
            highNibble = retCode
#
# bring enable down again
#
        cmd &= ~self.HD44780_EN
        retCode = pi.i2c_write_byte(self.handle,cmd)
#
# enable high again
#
        cmd |= self.HD44780_EN;
        retCode = pi.i2c_write_byte(self.handle,cmd)            
#
# read low nibble
#
        retCode = pi.i2c_read_byte(self.handle)
        if retCode < 0:
            print("in read: error when reading I2C bus")
            return
        else:
            lowNibble = retCode
#
# bring enable down again
#
        cmd &= ~self.HD44780_EN
        retCode = pi.i2c_write_byte(self.handle,cmd)
        return(highNibble & 0xf0) | lowNibble >> 4
#
# clear the display
#
    def clear(self):
        cmd = self.HD44780_CLEAR  
        retCode = self.__writeCmd(cmd)
        return retCode
#
# switch the display on or off
#
    def display(self,onOff):
        cmd = self.HD44780_DISPLAY_CURSOR
        if (onOff):
            cmd |= self.HD44780_DISPLAY_ON
        retCode = self.__writeCmd(cmd)
        return retCode
#
# home the cursor
#
    def home(self):
         cmd = self.HD44780_HOME
         retCode = self.__writeCmd(cmd)
         return retCode

#
# read the busy flag
#
    def readBusy(self):
        retCode = self.read(self.HD44780_CMD_REG)
        return retCode

#
# set up the address register
#
    def setDDramAddr(self,address):
        if  address > 80:
            print("setDDramAddr: addresses 0..80 allowed")
            return
        cmd = self.HD44780_SET_DDRAM | address
        self.__writeCmd(cmd)
    
#
# setting up the address register to write to the second line
#
    def secondLine(self):
        self.setDDramAddr(self.HD44780_2ND_LINE)

#
# define the entry mode
#
    def entry(self,incDec,shift):
        cmd = self.HD44780_ENTRY_MODE
#
# here for increment / decrement
#
        if incDec == 'inc':
#            print("in entry: increment")
            cmd |= self.HD44780_CURSOR_INC
        elif incDec == "dec":
            pass
#            print("in entry: decrement")
        else:
            print("in entry: increment/decrement mode ",incDec," is unknown. Skipping...")
#
# here for shift
#
        if shift == 'shift':
#            print("in entry: setting to shift mode")
            cmd |= self.HD44780_DISPLAY_SHIFT
        elif shift == '':
            pass
#            print("in entry: not shifting")
        else:
            print("in entry: shift mode ",shift," unknown. Skipping...")
        self.__writeCmd(cmd)
#
# cursor definitions
#
    def cursor(self,onOff,blink):
        cmd = self.HD44780_DISPLAY_CURSOR
#
# on or off
#
        if onOff == 'on':
            cmd |= self.HD44780_CURSOR_ON
        elif onOff == 'off':
            pass
        else:
            printf("in cursor: cursor mode ",onOff," unknown. Skipping...")
#
# blinking or steady
#
        if blink == 'blink':
            cmd |= self.HD44780_CURSOR_BLINK
        elif blink == 'steady':
            pass
        else:
            printf("in cursor: cursor blinking mode ",blink," unknown. Skipping...")
        self.__writeCmd(cmd)
        
#
# user defined character
#
    def userChar(self,address,dotMatrix):
#
# write the address of the cgram (character generator RAM)
#
        if address < 0 or address > 0x0f<<3:
            print("in userChar: illegal address. Skipping...")
            return
        if len(dotMatrix) != 7:
            print("in userChar: dot Matrix needs 7 dot lines. Not writing!")
            return

#
# read the current address of the ddram
#
        ddramAddress = self.readBusy() & 0x7f
# setup the cgram address
#        print("in userChar: ",format(address,'02x'))
        cmd = self.HD44780_SET_CGRAM | address
        self.__writeCmd(cmd)
# now we can write the dot matrix

        for index in range(0,7):
#            print("in userChar: line[",index,"] =0x"+format(dotMatrix[index],'02x'))
            self.putC(chr(dotMatrix[index]))
        cmd = self.HD44780_SET_DDRAM | ddramAddress
        self.__writeCmd(cmd)
                          
        
